/*
 * _UCODE_C_
 *
 * MICROCODE ENGINE FOR GC64-DMC ENGINE
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include "dmc_ucode.h"


/*
 * Globals
 *
 */

btree root;                   /* btree root */
btree *free_list;             /* free node list */
btree *free_list_last;        /* last free node in the list */
btree *nodes;                 /* node list */
int total_read_bytes  = 0;    /* total number of read bytes in the tree */
int total_write_bytes = 0;    /* total number of write bytes in the tree */
int read_timeout      = 0;    /* read timeout value */
int write_timeout     = 0;    /* write timeout value */
int num_nodes         = 0;    /* number of nodes in the tree */
int num_free          = 0;    /* number of nodes in the free list */
struct dmc __config;          /* dmc config struct */
hmc_list *head	      = NULL;
hmc_list *tail	      = NULL;
struct dmc __config;
uint64_t base_addr    = 0;    /* base address of the tree */ 
uint64_t temp_addr    = 0;    /* temporary address for the logic testing*/ 
btree *temp;                  /* temp node represent the previous node been inserted*/



/* ---------------------------------------- init the node of HMC LIST */
extern void hmc_init(hmc_list *node){
  node =(hmc_list*) malloc( sizeof( hmc_list ) );
  node->next=NULL;
}


/* ---------------------------------------- EXPIRE_TREE */
static void expire_tree(btree *p) {

  btree *new  = NULL;

  if(p == NULL) {
    return ;
  }

  /* descend into each subtree and put
   * all the nodes on the free list
   * first, get the end of the free
   * list
   */

  if( p->left != NULL ){
    expire_tree( p->left );
    p->left = NULL;
  }

  if( p->right != NULL ){
    expire_tree( p->right );
    p->right = NULL;
  }

  /*
   * put p back on the free list
   *
   */
  if( free_list == NULL ){
    free_list = p;
    free_list_last = p;
    num_free = 1;
    return ;
  }

  /*
   * existing nodes, get the last in the list
   *
   */
  new = free_list_last;

  /* insert the free node */
  new->left = p;
  p->left = NULL;
  free_list_last = p;
  num_free++;

}


/* ---------------------------------------- INIT_ROOT_NODE */
static void init_root_node(){
  root.left   = NULL;
  root.right  = NULL;
  root.parent = NULL;
}

/* ---------------------------------------- INIT_LEAF_NODE */
static void init_leaf_node(){
  int i =0;
  btree *tmp;

  /*
   * we only allocate the minimum number
   * of nodes; which is the maximum size
   * of a subtree or the maximum number
   * insertion cycles
   *
   */
  if( MAX_WRITE_TIME > 128 ){
    nodes = malloc( sizeof( btree ) * 256 );
    num_nodes = 256;
  }else{
    num_nodes = MAX_WRITE_TIME*2;
    nodes = malloc( sizeof( btree ) * num_nodes );
  }

#ifdef DEBUG
  printf( "num_nodes = %d\n", num_nodes );
#endif

  free_list = &(nodes[0]);
  tmp       = free_list;
  for( i=1; i<num_nodes; i++ ){
    tmp->left = &(nodes[i]);
    tmp = tmp->left;
  }

  nodes[num_nodes-1].left = NULL;
  free_list_last = &(nodes[num_nodes-1]);

  num_free = num_nodes;
}

/* ---------------------------------------- GET_LEFT_LEAF */
static btree *get_left_leaf( btree *src ){

  btree *tmp = NULL;

#ifdef DEBUG
  printf( "... in get_left_leaf\n" );
#endif

  /* if the left node is null,
   * then just return the src node
   * eg, no children present
   */
  if( src->left == NULL ){
    return src;
  }

  tmp = src->left;
  while( tmp->left != NULL ){
    tmp = tmp->left;
  }

  return tmp;
}

/* ---------------------------------------- GET_RIGHT_LEAF */
static btree *get_right_leaf( btree *src ){

  btree *tmp = NULL;

#ifdef DEBUG
  printf( "... in get_right_leaf\n" );
#endif

  /* if the right node is null,
   * then just return the src node
   * eg, no children present
   */
  if( src->right == NULL ){
    return src;
  }

  tmp = src->right;
  while( tmp->right != NULL ){
    tmp = tmp->right;
  }

  return tmp;
}
/* ---------------------------------------- GET_NEW_NODE */
static btree *get_new_node(){

  btree *tmp = free_list;
  btree *new = tmp;

#ifdef DEBUG
  printf( "... entering get_new_node\n" );
#endif

  if( num_free == 0 || free_list == NULL ){
    return NULL;
  }

  /* last node */
  if( free_list->left == NULL ){
    free_list = NULL;
    num_free  = 0;
    return tmp;
  }

  /* get the last node */
  while( new->left != NULL ){
    tmp = new;
    new = new->left;
  }

  num_free--;
  tmp->left   = NULL;
  new->left   = NULL;
  new->right  = NULL;

#ifdef DEBUG
  printf( "... exiting get_new_node\n" );
#endif

  return new;
}

/* ---------------------------------------- INSERT_CHILD */
static void insert_child( btree **parent, btree *child ){

#ifdef DEBUG
  printf( "...in insert_child\n" );
#endif

  /* if the parent is null, we're done */
  if( *parent == NULL ){
#ifdef DEBUG
    printf( "...found null parent\n" );
#endif
    *parent = child;
    return ;
  }

  /* recursively insert based upon the address */
  if( (*parent)->req.addr >= child->req.addr ){
    /* if new inserted addr is smaller
     * or equal to the parents' addr
     */
    if( (*parent)->left == NULL) {
      (*parent)->left = child;
      child->parent = (*parent);
      return ;
    }else{
      insert_child(&((*parent)->left),child);
      return ;
    }
  }else{
    /* if new inserted addr is larger
     * the parents' addr
     */
    if((*parent)->right == NULL) {
      (*parent)->right = child;
      child->parent = (*parent);
      return ;
    }else{
      insert_child(&((*parent)->right),child);
      return ;
    }
  }
}


/* ---------------------------------------- InOrderTraverse */
/* Read: io = 0
 * Write: io = 1
 */
static void InOrderTraverse(btree* n, int io, btree* end){

  if(n != NULL){

    InOrderTraverse(n->left, io,end);
    /* -------------------------- for the case total read bytes in the whole tree doesn't exceed 128*/
    if(n == end && io == 0){
	temp_addr = n->req.addr + n->req.op;
	total_read_bytes = temp_addr - base_addr;
	if(total_read_bytes <= 128){
		trace_read( total_read_bytes, base_addr );
	}
	return;
    }
    else if(n == end && io == 1){

	temp_addr = n->req.addr + (n->req.op&0xFF);
	total_write_bytes = temp_addr - base_addr;
	if(total_write_bytes <= 128){
		trace_write( total_write_bytes, base_addr );
	}
	return;
    }
    if( io == 0 ){
      /* READ */
      /* -------------------------- check if total read bytes exceed 128 or not*/
      if( (n->req.addr + n->req.op - base_addr < 128) ||
          (n->req.addr == base_addr) ){
        temp 	  = n;
        temp_addr = n->req.addr + n->req.op;
      }else{
        /* -------------------------- build one hmc request*/
        total_read_bytes = temp_addr - base_addr;
        trace_read( total_read_bytes, base_addr );

        /* -------------------------- reset the base address, temp node and temp address*/	
        temp 	     	  = n;
        base_addr    	  = n->req.addr;
        temp_addr    	  = n->req.addr + n->req.op;
        total_read_bytes    = 0;
      }

    }/* end read */
    else{
      /* WRITE */
      if( ((n->req.addr + (n->req.op&0xFF) - base_addr < 128) &&
          (n->req.addr == temp_addr)) ||
          (n->req.addr == base_addr)  ){
        temp 	  = n;
        temp_addr = n->req.addr + (n->req.op&0xFF);
      }else{
        /* -------------------------- build one hmc request*/
        total_write_bytes = temp_addr - base_addr;
        trace_write( total_write_bytes, base_addr );

        /* -------------------------- reset the base address, temp node and temp address*/	
        temp 	     	  = n;
        base_addr    	  = n->req.addr;
        temp_addr    	  = n->req.addr + (n->req.op&0xFF);
        total_write_bytes = 0;
      }
    }/* end write */

     InOrderTraverse(n->right, io,end);

  }
}


/******************************************************** insert HMC requset into the hmc list */
static void hmc_insert(int size, uint64_t addr, int flag ){
  hmc_list *node;
  hmc_init(node);
  int op;
  if( size%16 != 0 ){
    size = ((size/16)+1)*16;
    if( size>128){size=128;}
  }
  /* if flag = 0, the request is read operation*/
  if(flag == 0)
    op = size/16;
  /* if flag = 1, the request is write operation*/
  if(flag == 1)
    op = size/16+16;

  node->req.op = op;
  node->next = NULL;

  /* if the list is empty*/
  if( head == NULL )
    head = node;
  else
    tail->next = node;

  tail = node;

}

/* ---------------------------------------- expire the hmc_list */
extern void hmc_expire(){
   hmc_list *temp;
   hmc_list *node;
   temp =head;

   if(head->next == NULL){
	printf("the list is empty\n");
	return;
   }

   if(head->next == NULL){
	free(head);
	printf("the list is expired\n");
   	return;
   }

   while(temp != tail){
	node = temp->next;
	free(temp);
	temp=node;
   }

   free(tail);
   printf("the list is expired\n");
   return;
}

/* ---------------------------------------- UCODE_INSERT */
extern void ucode_insert(struct mtrace rqst){

  /*
   * call the tracing function before returning
   * the trace function handles the config
   * and only prints tracing if enabled
   *
   */
  btree* tmp  = NULL;
  btree* pri  = NULL;
  btree* end  = NULL;

#ifdef DEBUG
  printf( "... in ucode_insert\n" );
#endif

  /*
   * get a new node for the insertion
   *
   */
  tmp = get_new_node();
  if( tmp == NULL ){

    /* no new nodes, expire the trees */
#ifdef DEBUG
    printf("...out of nodes\n");
#endif

    /* -- read nodes */
    if( root.left != NULL ){
      pri = get_left_leaf( root.left );
      /*find out the most right nodes (the end of the left subtree)*/
      end = get_right_leaf( root.left );
      base_addr =  pri->req.addr;
      temp_addr =  0;
      temp      =  pri;
      InOrderTraverse(root.left,0,end);
      expire_tree(root.left);
      root.left = NULL;
      total_read_bytes  = 0;
      read_timeout      = 0;
    }

    /* -- write nodes */
    if( root.right != NULL ){
      pri = get_left_leaf( root.right );
      /*find out the most right nodes (the end of the right subtree)*/
      end = get_right_leaf( root.right );
      base_addr =  pri->req.addr;
      temp_addr =  0;
      temp      =  pri;
      InOrderTraverse(root.right,1,end);
      expire_tree(root.right);
      root.right = NULL;
      total_write_bytes = 0;
      write_timeout     = 0;
    }

    /* get a new node */
    tmp = get_new_node();
  }

#ifdef DEBUG
  printf( "...got a new node\n" );
#endif

  /* fill in the data */
  tmp->req.proc = rqst.proc;
  tmp->req.op   = rqst.op;
  tmp->req.addr = rqst.addr;

  /* insert read requests from RISCV cores */
  if( (tmp->req.op >= RD1) && (tmp->req.op <= RD16) ) {
#ifdef DEBUG
    printf( "...inserting read request\n" );
#endif
    insert_child( &root.left, tmp );
    total_read_bytes+=tmp->req.op;
    read_timeout++;
  }

  /* insert write requests from RISCV cores */
  if( (tmp->req.op >= RD16) && (tmp->req.op <= WR16) ) {
#ifdef DEBUG
    printf( "...inserting write request\n" );
#endif
    insert_child( &root.right, tmp );
    total_write_bytes+=(tmp->req.op&0xFF);
    write_timeout++;
  }

#ifdef DEBUG
  printf( "...request inserted\n" );
#endif

  /*
   * force an HMC read request
   * set timeout to zero
   * clear read requests in the tree
   *
   */
  if( (total_read_bytes == 128) || (read_timeout >= MAX_READ_TIME) ){
#ifdef DEBUG
    printf( "...expiring read tree\n" );
#endif
    pri = get_left_leaf( root.left );
    end = get_right_leaf( root.left );
    /* --- add logic to build the hmc requests. */

    base_addr =  pri->req.addr;
    temp_addr =  0;
    temp      =  pri;

    /* --- traverse the tree and build hmc requests */
    InOrderTraverse(root.left,0,end);

    expire_tree(root.left);
    root.left         = NULL;

    total_read_bytes  = 0;
    read_timeout      = 0;
  }

  /*
   * force an HMC write request
   * set timeout to zero
   * clear the write requests in the tree
   */
  if( (total_write_bytes == 128) || (write_timeout >= MAX_WRITE_TIME) ){
#ifdef DEBUG
    printf( "...expiring write tree\n" );
#endif
    pri = get_left_leaf( root.right );
    end = get_right_leaf( root.right );

    base_addr =  pri->req.addr;
    temp_addr =  0;
    temp      =  pri;

    InOrderTraverse(root.right,1,end);

    expire_tree(root.right);
    root.right        = NULL;

    total_write_bytes = 0;
    write_timeout     = 0;
  }

  trace();
}


/* ---------------------------------------- UCODE_FLUSH */
extern void ucode_flush(){

  btree* pri  = NULL;
  btree* end  = NULL;

  /* -- read nodes */
#ifdef DEBUG
  printf( "...ucode_flush()\n" );
  printf( "...flushing read nodes\n" );
#endif
  if( root.left != NULL ){
    pri = get_left_leaf( root.left );
    end = get_right_leaf( root.left );
    base_addr =  pri->req.addr;
    temp_addr =  0;
    temp      =  pri;
    InOrderTraverse(root.left,0,end);
    expire_tree(root.left);
    root.left = NULL;
    total_read_bytes  = 0;
    read_timeout      = 0;
  }

  /* -- write nodes */
#ifdef DEBUG
  printf( "...flushing write nodes\n" );
#endif
  if( root.right != NULL ){
    pri = get_left_leaf( root.right );
    end = get_right_leaf( root.right );
    base_addr =  pri->req.addr;
    temp_addr =  0;
    temp      =  pri;
    InOrderTraverse(root.right,1,end);
    expire_tree(root.right);
    root.right = NULL;
    total_write_bytes = 0;
    write_timeout     = 0;
  }

}

/* ---------------------------------------- UCODE_FREE */
extern void ucode_free() {
  free(nodes);
  nodes = NULL;
}

/* ---------------------------------------- UCODE_INIT */
extern void ucode_init( uint64_t conf ) {

  /* init the structure */
  __config.conf = conf;

  /* init the root node */
  init_root_node();

  /* init the leaf nodes */
  init_leaf_node();
}

/* EOF */
